unit ZREval;

{$B-}

interface

uses
  Windows, WinSpool, Sysutils, Messages, Classes, Controls, StdCtrls, ExtCtrls,
  ComCtrls, Buttons, Printers, Graphics, Forms, Dialogs, DB, DBTables, ZRPrntr;

type
  { TZRLibraryEntry }
  TZRLibraryItemClass = class of TObject;

  TZRLibraryEntry = class
    private
      fName        : String;
      fDescription : String;
      fVendor      : String;
      fItem        : TZRLibraryItemClass;
    public
      property Description : String read fDescription write fDescription;
      property Name        : String read fName write fName;
      property Vendor      : String read fVendor write fVendor;
      property Item        : TZRLibraryItemClass read fItem write fItem;
  end;

  { TZRLibrary }
  TZRLibrary = class
    private
      fEntries : TStrings;
    protected
      function GetEntry(Index : integer) : TZRLibraryEntry; virtual;
    public
      constructor Create;
      destructor  Destroy; override;
      procedure Add(aItem : TZRLibraryItemClass; aName, aDescription, aVendor: String);
      property Entries : TStrings read fEntries write fEntries;
      property Entry[Index : integer] : TZRLibraryEntry read GetEntry; defaut;
  end;

  TZREvElement = class;

  { TZRFunctionLibrary }
  TZRFunctionLibrary = class(TZRLibrary)
    public
      function GetFunction(Name : String) : TZREvElement;
  end;

  { TZREvaluator related declarations }
  TZREvOperator = (opLess, opLessOrEqual, opGreater, opGreaterOrEqual, opEqual,
                   opUnequal, opPlus, opMinus, opOr, opMul, opDiv, opMod, opAnd);

  TZREvValueKind = (valInteger, valDouble, valString, valBoolean, valDate, valError);

  TZREvValueType = record
    case Kind : TZREvValueKind of
      valInteger: (intResult : Longint);
      valDouble : (dblResult : Double);
      valString : (strResult : String[255]);
      valDate   : (datResult : TDateTime);
      valBool   : (booResult : Boolean);
  end;

  TZREvResult = class
    public
      Value : TZREvValueType;
  end;

  TZRFiFo = class
    private
      fAggreg   : Boolean;
      fFiFo     : TList;
      fNextItem : Integer;
    public
      constructor Create;
      destructor  Destroy; override;
      procedure Start;
      procedure Put(Value : TObject);
      function  Get        : TObject;
      property Aggreg : Boolean read fAggreg write fAggreg;
  end;

  TZREvElement = class
    private
      fIsAggreg : Boolean;
    public
      constructor Create; virtual;
      function  Value(FiFo : TZRFiFo) : TZREvResult; virtual;
      procedure Reset; virtual;
      property IsAggreg : Boolean read fIsAggreg write fIsAggreg;
  end;

  TZREvElementFunction = class(TZREvElement)
    private
    protected
      ArgList : TList;
      function  ArgumentOK(Value : TZREvElement) : Boolean;
      function  Argument(Index : Integer) : TZREvResult;
      procedure FreeArguments;
      procedure GetArguments(FiFo : TZRFiFo);
      procedure Aggregate; virtual;
      function  Calculate : TZREvResult; virtual;
    public
      constructor Create; override;
      destructor Destroy; override;
      function Value(FiFo : TZRFiFo) : TZREvResult; override;
  end;

  TZREvElementFunctionClass = class of TZREvElementFunction;

  TZREvElementArgumentEnd = class(TZREvElement);

  TZREvElementDataField = class(TZREvElement)
    private
      fDataSet : TDataSet;
      fFieldNo : Integer;
      fField   : TField;
    public
      constructor CreateField(aField : TField); virtual;
      function Value(FiFo : TZRFiFo) : TZREvResult; override;
  end;

  TZREvElementError = class(TZREvElement)
    public
      function Value(FiFo : TZRFiFo) : TZREvResult; override;
  end;

  { TZREvaluator class }

  TZREvaluator = class (TObject)
    private
      fDataSets  : TList;
      FiFo       : TZRFiFo;
      fPrepared  : Boolean;
      function  EvalFunctionExpr(const strFunc : String) : TZREvResult;
      function  EvalSimpleExpr(const strSimplExpr : String) : TZREvResult;
      function  EvalTerm(const strTermExpr : String) : TZREvResult;
      function  EvalFactor(strFactorExpr : String) : TZREvResult;
      function  EvalString(const strString : String) : TZREvResult;
      function  EvalConstant(const strConstant : String) : TZREvResult;
      function  GetAggregate : Boolean;
      function  NegateResult(const Res : TZREvResult) : TZREvResult;
      function  Evaluate(const strExpr : String) : TZREvResult;
      procedure FindDelimiter(strArg : String; var Pos : Integer);
      procedure FindOp1(const strExpr : String; var Op : TZREvOperator; var Pos, Len : Integer);
      procedure FindOp2(const strExpr : String; var Op : TZREvOperator; var Pos, Len : Integer);
      procedure FindOp3(const strExpr : String; var Op : TZREvOperator; var Pos, Len : Integer);
      procedure SetAggregate(Value : Boolean);
      procedure TrimString(var strString : String);
    protected
      function EvalFunction(strFunc : String; const strArg : String) : TZREvResult; virtual;
      function EvalVariable(strVariable : String) : TZREvResult; virtual;
      function GetIsAggreg : Boolean;
    public
      constructor Create;
      destructor Destroy; override;
      function  Calculate(const StrExpr : String) : TZREvResult;
      function  Value : TZREvResult;
      procedure Prepare(const StrExpr : String);
      procedure Reset;
      procedure UnPrepare;
      property IsAggreg  : Boolean read GetIsAggreg;
      property Aggregate : Boolean read GetAggregate write SetAggregate;
      property DataSets  : TList   read fDataSets write fDataSets;
      property Prepared  : Boolean read fPrepared write fPrepared;
  end;

function Calculate(Formula: String; Values: array of const): Variant;

procedure RegisterZRFunction(FunctionClass : TZRLibraryItemClass; Name, Description, Vendor, Arguments : String);

var
  ZRFunctionLibrary : TZRFunctionLibrary;

implementation

uses
  ZRStrUtl, ZRConst, ZReport;

function Calculate(Formula: String; Values: array of const): Variant;
var
  E: TZREvaluator;
  R: TZREvResult;
begin
  Formula:= Format(Formula, Values);
  E:= TZREvaluator.Create;
  try
    R:= E.Calculate(Formula);
    case R.Kind of
      resInteger: Result:= R.intResult;
      resDouble : Result:= R.dblResult;
      resString : Result:= R.strResult;
      resBool   : Result:= R.booResult;
      else        Result:= Null;
    end;
  except
    E.Free;
    VarClear(Result);
  end;
end;


const
  ArgSeparator : Char = ',';

procedure RegisterZRFunction(FunctionClass : TZRLibraryItemClass; Name, Description, Vendor, Arguments : string);
begin
  ZRFunctionLibrary.Add(FunctionClass, Name, Description, Vendor, Arguments);
end;

{ TZRLibrary }

constructor TZRLibrary.Create;
begin
  inherited Create;
  fEntries := TStringList.Create;
end;

destructor TZRLibrary.Destroy;
var
  I : integer;
begin
  for I := 0 to Entries.Count - 1 do
    fEntries.Objects[I].Free;
  fEntries.Free;
  inherited Destroy;
end;

procedure TZRLibrary.Add(aItem : TZRLibraryItemClass; aName, aDescription, aVendor: String);
var
  aEntry : TZRLibraryEntry;
begin
  aEntry := TZRLibraryEntry.Create;
  with aEntry do
    begin
      Name        := aName;
      Description := aDescription;
      Vendor      := aVendor;
      Item        := aItem;
    end;
  fEntries.AddObject(aName, aEntry);
end;

function TZRLibrary.GetEntry(Index : integer) : TZRLibraryEntry;
begin
  if Index <= fEntries.Count then
    Result := TZRLibraryEntry(fEntries.Objects[Index])
  else
    Result := nil;
end;

{ TZRFunctionLibrary }

function TZRFunctionLibrary.GetFunction(Name : String) : TZREvElement;
var
  I             : integer;
  AObject       : TZREvElementFunctionClass;
  aLibraryEntry : TZRLibraryEntry;
begin
  I := Entries.IndexOf(Name);
  if I>=0 then
    begin
      aLibraryEntry := TZRLibraryEntry(Entry[I]);
      aObject := TZREvElementFunctionClass(aLibraryEntry.Item);
      Result := aObject.Create;
    end
  else
    Result := TZREvElementError.Create;
end;

{ TZREvaluator }

constructor TZRFiFo.Create;
begin
  fFiFo     := TList.Create;
  fAggreg   := False;
  fNextItem := 0;
end;

destructor TZRFiFo.Destroy;
var
  I : integer;
begin
  for I := 0 to FiFo.Count-1 do
    TObject(fFiFo[I]).Free;
  fFiFo.Free;
  inherited Destroy;
end;

procedure TZRFiFo.Start;
begin
  fNextItem := 0;
end;

procedure TZRFiFo.Put(Value : TObject);
begin
  fFiFo.Add(Value);
end;

function TZRFiFo.Get : TObject;
begin
  if fNextItem < fFiFo.Count then
  begin
    Result := fFiFo[fNextItem];
    Inc(fNextItem);
  end else
    Result := nil;
end;

{ TZREvElement }

constructor TZREvElement.Create;
begin
  inherited Create;
  fIsAggreg := False;
end;

function TZREvElement.Value(FiFo : TZRFiFo) : TZREvResult;
begin
end;

procedure TZREvElement.Reset;
begin
end;

{ TZREvElementOperator }

type
  TZREvElementOperator = class(TZREvElement)
  private
    fCode :  TZREvOperator;
    procedure ConvertZREvResults(var Res1 : TZREvResult; var Res2 : TZREvResult);
  public
    constructor CreateOperator(Code : TZREvOperator);
    function Value(FiFo : TZRFiFo) : TZREvResult; override;
  end;

constructor TZREvElementOperator.CreateOperator(Code : TZREvOperator);
begin
  inherited Create;
  fCode := Code;
end;

procedure TZREvElementOperator.ConvertZREvResults(var Res1 : TZREvResult; var Res2 : TZREvResult);
begin
  if Res1.Kind <> Res2.Kind then
    begin
      if ((Res1.Kind = resInteger) and (Res2.Kind = resDouble)) then
        begin
          Res1.Kind      := resDouble;
          Res1.dblResult := Res1.intResult;
        end else
      if ((Res1.Kind = resDouble) and (Res2.Kind = resInteger)) then
        begin
          Res2.Kind      := resDouble;
          Res2.dblResult := Res2.intResult;
        end
      else
        begin
          Res1.Kind := resError;
          Res2.Kind := resError;
        end;
    end;
end;

function TZREvElementOperator.Value(FiFo : TZRFiFo) : TZREvResult;
var
  Res1,
  Res2 : TZREvResult;
begin
  Res1 := TZREvElement(FiFo.Get).Value(FiFo);
  Res2 := TZREvElement(FiFo.Get).Value(FiFo);
  ConverTZREvResults(Res1, Res2);
  Result.Kind := Res1.Kind;
  if Result.Kind <> resError then
    case fCode of
      opPlus: case result.Kind of
                resInteger: result.intResult := Res1.intResult + Res2.intResult;
                resDouble : result.dblResult := Res1.dblResult + Res2.dblResult;
                resDate   : result.datResult := Res1.datResult + Res2.datResult;
                resString : result.strResult := Res1.strResult + Res2.strResult;
                resBoolean: result.Kind := resError;
              end;
      opMinus:  case result.Kind of
                  resInteger: result.intResult := Res1.intResult - Res2.intResult;
                  resDouble : result.dblResult := Res1.dblResult - Res2.dblResult;
                  resDate   : result.datResult := Res1.datResult - Res2.datResult;
                  resString : result.Kind := resError;
                  resBoolean: result.Kind := resError;
                end;
      opMul: case result.Kind of
               resInteger: result.intResult := Res1.intResult * Res2.intResult;
               resDouble : result.dblResult := Res1.dblResult * Res2.dblResult;
               resString : result.Kind := resError;
               resDate   : Result.Kind := resError;
               resBoolean: result.Kind := resError;
             end;
      opMod: case result.Kind of
               resInteger: if Res2.intResult <> 0 then
                             Result.intResult := Res1.intResult mod Res2.intResult;
                           else
                             Result.Kind := resError;
               resDouble : Result.Kind := resError;
               resString : Result.Kind := resError;
               resDate   : Result.Kind := resError;
               resBoolean: Result.Kind := resError;
             end;
      opDiv: case result.Kind of
               resInteger: if Res2.intResult <> 0 then
                             Result.intResult := Res1.intResult div Res2.intResult;
                           else
                             Result.Kind := resError;
               resDouble : if Res2.dblResult <> 0 then
                             Result.dblResult := Res1.dblResult / Res2.dblResult
                           else
                             Result.Kind := resError;
               resDate   : Result.Kind := resError;
               resString : Result.Kind := resError;
               resBoolean: Result.Kind := resError;
             end;
      opGreater: begin
                   result.Kind := resBool;
                   case Res1.Kind of
                     resInteger: result.booResult := Res1.intResult > Res2.intResult;
                     resDouble : result.booResult := Res1.dblResult > Res2.dblResult;
                     resString : result.booResult := Res1.strResult > Res2.strResult;
                     resDate   : result.booResult := Res1.datResult > Res2.datResult;
                     resBoolean: result.Kind := resError;
                   end;
                 end;
      opGreaterOrEqual: begin
                result.Kind := resBool;
                case Res1.Kind of
                  resInteger: result.booResult := Res1.intResult >= Res2.intResult;
                  resDouble : result.booResult := Res1.dblResult >= Res2.dblResult;
                  resString : result.booResult := Res1.strResult >= Res2.strResult;
                  resDate   : result.booResult := Res1.datResult >= Res2.datResult;
                  resBoolean: result.Kind := resError;
                end;
             end;
      opLess: begin
                result.Kind := resBool;
                case Res1.Kind of
                  resInteger: result.booResult := Res1.intResult < Res2.intResult;
                  resDouble : result.booResult := Res1.dblResult < Res2.dblResult;
                  resString : result.booResult := Res1.strResult < Res2.strResult;
                  resDate   : result.booResult := Res1.datResult < Res2.datResult;
                  resBoolean: result.Kind := resError;
                end;
              end;
      opLessOrEqual: begin
                 result.Kind := resBool;
                 case Res1.Kind of
                   resInteger: result.booResult := Res1.intResult <= Res2.intResult;
                   resDouble : result.booResult := Res1.dblResult <= Res2.dblResult;
                   resString : result.booResult := Res1.strResult <= Res2.strResult;
                   resDate   : result.booResult := Res1.datResult <= Res2.datResult;
                   resBoolean: result.Kind := resError;
                 end;
               end;
      opEqual: begin
                 result.Kind := resBool;
                 case Res1.Kind of
                   resInteger: result.booResult := Res1.intResult = Res2.intResult;
                   resDouble : result.booResult := Res1.dblResult = Res2.dblResult;
                   resString : result.booResult := Res1.strResult = Res2.strResult;
                   resBoolean: result.booResult := Res1.booResult = Res2.booResult;
                   resDate   : result.booResult := Res1.datResult = Res2.datResult;
                 end;
               end;
      opUnequal: begin
                   result.Kind := resBool;
                   case Res1.Kind of
                     resInteger: result.booResult := Res1.intResult <> Res2.intResult;
                     resDouble : result.booResult := Res1.dblResult <> Res2.dblResult;
                     resString : result.booResult := Res1.strResult <> Res2.strResult;
                     resBoolean: result.booResult := Res1.booResult <> Res2.booResult;
                     resDate   : result.booResult := Res1.datResult <> Res2.datResult;
                   end;
                 end;
      opOr: begin
              result.Kind := Res1.Kind;
              case Res1.Kind of
                resInteger: result.intResult := Res1.intResult or Res2.intResult;
                resDouble : result.Kind := resError;
                resDate   : result.Kind := resError;
                resString : Result.Kind := resError;
                resBoolean: result.booResult := Res1.booResult or Res2.booResult;
              end;
            end;
      opAnd: begin
               result.Kind := Res1.Kind;
               case Res1.Kind of
                 resInteger: result.intResult := Res1.intResult and Res2.intResult;
                 resDouble : result.Kind := resError;
                 resDate   : result.Kind := resError;
                 resString : result.Kind := resError;
                 resBoolean: result.booResult := Res1.booResult and Res2.booResult;
               end;
             end;
    end;
end;

{ TZREvElementConstant }

type
  TZREvElementConstant = class(TZREvElement)
  private
    fValue : TZREvResult;
  public
    constructor CreateConstant(Value : TZREvResult);
    function Value(FiFo : TZRFiFo) : TZREvResult; override;
  end;

constructor TZREvElementConstant.CreateConstant(Value : TZREvresult);
begin
  inherited Create;
  fValue := Value;
end;

function TZREvElementConstant.Value(FiFo : TZRFiFo): TZREvResult;
begin
  Result := fValue;
end;

{ TZREvElementString }

type
  TZREvElementString = class(TZREvElement)
  private
    fValue : string;
  public
    constructor CreateString(Value : string);
    function Value(FiFo : TZRFiFo) : TZREvResult; override;
  end;

constructor TZREvElementString.CreateString(Value : string);
begin
  inherited Create;
  fValue := Value;
end;

function TZREvElementString.Value(FiFo : TZRFiFo) : TZREvResult;
begin
  Result.Kind      := resString;
  Result.StrResult := fValue;
end;

{ TZREvElementFunction }

constructor TZREvElementFunction.Create;
begin
  inherited Create;
  ArgList := TList.Create;
end;

destructor TZREvElementFunction.Destroy;
begin
  ArgList.Free;
  inherited Destroy;
end;

procedure TZREvElementFunction.GetArguments(FiFo : TZRFiFo);
var
  aArgument : TZREvElement;
  AResult : TZREvResultClass;
begin
  repeat
    aArgument := TZREvElement(FiFo.Get);
    if not (aArgument is TZREvElementArgumentEnd) then
    begin
      aResult := TZREvResultClass.Create;
      aResult.EvResult := aArgument.Value(FiFo);
      ArgList.Add(aResult);
    end;
  until aArgument is TZREvElementArgumentEnd;
end;

procedure TZREvElementFunction.FreeArguments;
var
  I : integer;
begin
  for I := 0 to ArgList.Count - 1 do
    TZREvElement(ArgList.Items[I]).Free;
  ArgList.Clear;
end;

function TZREvElementFunction.Argument(Index : integer): TZREvResult;
begin
  if Index <= ArgList.Count then
    Result := TZREvResultClass(ArgList[Index]).EvResult;
end;

function TZREvElementFunction.Value(FiFo : TZRFiFo) : TZREvResult;
begin
  GetArguments(FiFo);
  if FiFo.Aggreg then
    Aggregate;
  Result := Calculate;
  FreeArguments;
end;

function TZREvElementFunction.ArgumentOK(Value : TZREvElement) : boolean;
begin
  Result := not (Value is TZREvElementArgumentEnd) and not (Value is TZREvElementError);
end;

procedure TZREvElementFunction.Aggregate;
begin
end;

function TZREvElementFunction.Calculate : TZREvResult;
begin
  Result.Kind := resError;
end;

{ TZREvTrue }

type
  TZREvTrue = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

  TZREvFalse = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

function TZREvTrue.Calculate : TZREvResult;
begin
  if ArgList.Count = 0 then
  begin
    Result.Kind := resBool;
    Result.booResult := true;
  end else
    Result.Kind := resError;
end;

function TZREvFalse.Calculate : TZREvResult;
begin
  if ArgList.Count = 0 then
  begin
    Result.Kind := resBool;
    Result.booResult := false;
  end else
    Result.Kind := resError;
end;

{ TZREvIfFunction }

type
  TZREvIfFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

function TZREvIfFunction.Calculate : TZREvResult;
begin
  if (ArgList.Count = 3) and (Argument(0).Kind = resBool) then
  begin
    if Argument(0).BooResult then
      Result := Argument(1)
    else
      Result := Argument(2);
  end else
    result.Kind := resError;
end;

{ TZREvTypeOfFunction }

type
  TZREvTypeOfFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

function TZREvTypeOfFunction.Calculate : TZREvResult;
begin
  Result.Kind := resString;
  if ArgList.Count = 1 then
  begin
    case Argument(0).Kind of
      resInteger: Result.StrResult := 'INTEGER';    {<-- do not resource }
      resDouble : Result.StrResult := 'FLOAT';   {<-- do not resource }
      resString : Result.StrResult := 'STRING';  {<-- do not resource }
      resBoolean: Result.StrResult := 'BOOLEAN';   {<-- do not resource }
      resDate   : Result.StrResult := 'DATE';   {<-- do not resource }
      resError  : Result.StrResult := 'ERROR';    {<-- do not resource }
    else
      Result.Kind := resError;
    end
  end else
    Result.Kind := resError;
end;

{ TZREvIntFunction }

type
  TZREvIntFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

function TZREvIntFunction.Calculate : TZREvResult;
begin
  Result.Kind := resInt;
  if ArgList.Count = 1 then
  begin
    case Argument(0).Kind of
      resInt : Result.IntResult := Argument(0).IntResult;
      resDouble : Result.IntResult := Round(Int(Argument(0).DblResult));
    else
      Result.Kind := resError;
    end
  end else
    Result.Kind := resError;
end;

{ TZREvFracFunction }

type
  TZREvFracFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

function TZREvFracFunction.Calculate : TZREvResult;
begin
  Result.Kind := resDouble;
  if ArgList.Count = 1 then
  begin
    case Argument(0).Kind of
      resInt : Result.DblResult := 0;
      resDouble : Result.DblResult := Frac(Argument(0).DblResult);
    else
      Result.Kind := resError;
    end
  end else
    Result.Kind := resError;
end;

{ TZREvSQRTFunction }

type
  TZREvSQRTFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

function TZREvSQRTFunction.Calculate : TZREvResult;
begin
  Result.Kind := resDouble;
  if ArgList.Count = 1 then
  begin
    case Argument(0).Kind of
      resInt : Result.DblResult := SQRT(Argument(0).IntResult);
      resDouble : Result.DblResult := SQRT(Argument(0).DblResult);
    else
      Result.Kind := resError;
    end
  end else
    Result.Kind := resError;
end;

{ TZREvStrFunction }

type
  TZREvStrFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

function TZREvStrFunction.Calculate : TZREvResult;
var
  ArgRes : TZREvResult;
begin
  if ArgList.Count = 1 then
  begin
    ArgRes := Argument(0);
    Result.Kind := resString;
    case ArgRes.Kind of
      resInt : Result.strResult := IntToStr(ArgRes.IntResult);
      resDouble : Result.strResult := FloatToStr(ArgRes.DblResult);
      resBool : if ArgRes.booResult then
                result.StrResult := 'True'          {<-- do not resource }
              else
                result.StrResult := 'False';        {<-- do not resource }
    else
      result.Kind := resError;
    end
  end else
    Result.Kind := resError;
end;

{ TZREvUpperFunction }

type
  TZREvUpperFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

function TZREvUpperFunction.Calculate : TZREvResult;
begin
  if (ArgList.Count = 1) and (Argument(0).Kind = resString) then
  begin
    Result.Kind := resString;
    Result.StrResult := AnsiUpperCase(Argument(0).StrResult);
  end else
    Result.Kind := resError;
end;

{ TZREvLowerFunction }

type
  TZREvLowerFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

function TZREvLowerFunction.Calculate : TZREvResult;
begin
  if (ArgList.Count = 1) and (Argument(0).Kind = resString) then
  begin
    Result.Kind := resString;
    Result.StrResult := AnsiLowerCase(Argument(0).StrResult);
  end else
    Result.Kind := resError;
end;

{ TZREvUpperFunction }

type
  TZREvPrettyFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

function TZREvPrettyFunction.Calculate : TZREvResult;
begin
  if (ArgList.Count = 1) and (Argument(0).Kind = resString) then
  begin
    Result.Kind := resString;
    Result.StrResult := AnsiUpperCase(Copy(Argument(0).StrResult, 1, 1)) +
                        AnsiLowerCase(Copy(Argument(0).StrResult, 2, length(Argument(0).StrResult)));
  end else
    Result.Kind := resError;
end;

{ TZREvTimeFunction }

type
  TZREvTimeFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

function TZREvTimeFunction.Calculate : TZREvResult;
begin
  if ArgList.Count = 0 then
  begin
    Result.Kind := resString;
    Result.StrResult := TimeToStr(Now);
  end else
    Result.Kind := resError;
end;

{ TZREvTimeFunction }

type
  TZREvDateFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

function TZREvDateFunction.Calculate : TZREvResult;
begin
  if ArgList.Count=0 then
  begin
    Result.Kind := resString;
    Result.StrResult := DateToStr(Date);
  end else
    Result.Kind := resError;
end;

{ TZREvCopyFunction }

type
  TZREvCopyFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

function TZREvCopyFunction.Calculate : TZREvResult;
begin
  if (ArgList.Count = 3) and
    (Argument(0).Kind = resString) and
    (Argument(1).Kind = resInt) and
    (Argument(2).Kind = resInt) then
  begin
    Result.StrResult := copy(Argument(0).strResult, Argument(1).IntResult, Argument(2).IntResult);
    Result.Kind := resString;
  end else
    Result.Kind := resError;
end;

{ TZREvFormatNumericFunction }

type
  TZREvFormatNumericFunction = class(TZREvElementFunction)
  public
    function Calculate : TZREvResult; override;
  end;

function TZREvFormatNumericFunction.Calculate : TZREvResult;
begin
  if (ArgList.Count = 2) and
     ((Argument(1).Kind = resInt) or (Argument(1).Kind = resDouble)) and
     (Argument(0).Kind = resString) then
  begin
    Result.Kind := resString;
    if Argument(1).Kind = resInt then
      Result.StrResult := FormatFloat(Argument(0).StrResult, Argument(1).IntResult * 1.0)
    else
      Result.StrResult := FormatFloat(Argument(0).StrResult, Argument(1).DblResult);
  end else
    Result.Kind := resError;
end;

{ TZREvSumFunction }

type
  TZREvSumFunction = class(TZREvElementFunction)
  private
    SumResult : TZREvResult;
    ResAssigned : boolean;
  public
    constructor Create; override;
    procedure Aggregate; override;
    function Calculate : TZREvResult; override;
    procedure Reset; override;
  end;

constructor TZREvSumFunction.Create;
begin
  inherited Create;
  ResAssigned := false;
  IsAggreg := true;
end;

procedure TZREvSumFunction.Reset;
begin
  ResAssigned := false;
end;

procedure TZREvSumFunction.Aggregate;
var
  aValue : TZREvResult;
begin
  if ArgList.Count = 1 then
  begin
    aValue := Argument(0);
    if ResAssigned then
    begin
      case aValue.Kind of
        resInt : SumResult.IntResult := SumResult.IntResult + aValue.IntResult;
        resDouble : SumResult.dblResult := SumResult.dblResult + aValue.dblResult;
        resString : SumResult.Kind := resError;
      end;
    end else
    begin
      SumResult.Kind := aValue.Kind;
      case aValue.Kind of
        resInt : SumResult.IntResult := aValue.IntResult;
        resDouble : SumResult.dblResult := aValue.dblResult;
      else
        SumResult.Kind := resError;
      end;
    end;
    ResAssigned := true;
  end else
    SumResult.Kind := resError;
end;

function TZREvSumFunction.Calculate : TZREvResult;
begin
  if ResAssigned then
    Result := SumResult
  else
    Result.Kind := resError;
end;

{ TZREvAverageFunction }

type
  TZREvAverageFunction = class(TZREvSumFunction)
  private
    Count : longint;
    aResult : TZREvResult;
  public
    function Calculate : TZREvResult; override;
    procedure Aggregate; override;
    procedure Reset; override;
  end;

procedure TZREvAverageFunction.Reset;
begin
  inherited Reset;
  aResult.Kind := resError;
  Count := 0;
  IsAggreg := true;
end;

procedure TZREvAverageFunction.Aggregate;
var
  aValue : TZREvResult;
begin
  inherited Aggregate;
  inc(Count);
  aValue := inherited Calculate;
  aResult.Kind := resDouble;
  case aValue.Kind of
    ResInt : aResult.DblResult := aValue.IntResult / Count;
    ResDouble : aResult.DblResult := aValue.DblResult / Count;
  else
    aResult.Kind := resError;
  end;
end;

function TZREvAverageFunction.Calculate : TZREvResult;
begin
  Result := aResult
end;

{ TZREvMaxFunction }

type
  TZREvMaxFunction = class(TZREvElementFunction)
  private
    MaxResult : TZREvResult;
    ResAssigned : boolean;
  public
    constructor Create; override;
    function Calculate : TZREvResult; override;
    procedure Aggregate; override;
    procedure Reset; override;
  end;

constructor TZREvMaxFunction.Create;
begin
  inherited Create;
  ResAssigned := false;
  IsAggreg := true;
end;

procedure TZREvMaxFunction.Reset;
begin
  ResAssigned := false;
end;

procedure TZREvMaxFunction.Aggregate;
var
  aValue : TZREvResult;
begin
  if ArgList.Count = 1 then
  begin
    aValue := Argument(0);
    if ResAssigned then
    begin
      case MaxResult.Kind of
        resInt : if aValue.IntResult > MaxResult.IntResult then
                   MaxResult.IntResult := aValue.IntResult;
        resDouble : if aValue.DblResult > MaxResult.DblResult then
                      MaxResult.DblResult := aValue.DblResult;
        resString : if aValue.StrResult > MaxResult.StrResult then
                      MaxResult.StrResult := aValue.StrResult;
        resBool : if aValue.booResult > MaxResult.BooResult then
                    MaxResult.BooResult := aValue.BooResult;
      else
        MaxResult.Kind := resError;
      end
    end else
    begin
      MaxResult := aValue;
      ResAssigned := true;
    end
  end else
    MaxResult.Kind := resError;
end;

function TZREvMaxFunction.Calculate : TZREvResult;
begin
  if ResAssigned then
    Result := MaxResult
  else
    Result.Kind := resError;
end;

{ TZREvMinFunction }

type
  TZREvMinFunction = class(TZREvElementFunction)
  private
    MinResult : TZREvResult;
    ResAssigned : boolean;
  public
    constructor Create; override;
    function Calculate : TZREvResult; override;
    procedure Aggregate; override;
    procedure Reset; override;
  end;

constructor TZREvMinFunction.Create;
begin
  inherited Create;
  ResAssigned := false;
  IsAggreg := true;
end;

procedure TZREvMinFunction.Reset;
begin
  ResAssigned := false;
end;

procedure TZREvMinFunction.Aggregate;
var
  aValue : TZREvResult;
begin
  if ArgList.Count = 1 then
  begin
    aValue := Argument(0);
    if ResAssigned then
    begin
      case MinResult.Kind of
        resInt : if aValue.IntResult < MinResult.IntResult then
                   MinResult.IntResult := aValue.IntResult;
        resDouble : if aValue.DblResult < MinResult.DblResult then
                      MinResult.DblResult := aValue.DblResult;
        resString : if aValue.StrResult < MinResult.StrResult then
                      MinResult.StrResult := aValue.StrResult;
        resBool : if aValue.booResult > MinResult.BooResult then
                    MinResult.BooResult := aValue.BooResult;
      else
        MinResult.Kind := resError;
      end
    end else
    begin
      MinResult := aValue;
      ResAssigned := true;
    end
  end else
    MinResult.Kind := resError;
end;

function TZREvMinFunction.Calculate : TZREvResult;
begin
  if ResAssigned then
    Result := MinResult
  else
    Result.Kind := resError;
end;

{ TZREvCountFunction }

type
  TZREvCountFunction = class(TZREvElementFunction)
  private
    FCount : integer;
  public
    constructor Create; override;
    function Value(FiFo : TZRFiFo) : TZREvResult; override;
    procedure Reset; override;
  end;

constructor TZREvCountFunction.Create;
begin
  inherited Create;
  FCount := 0;
  IsAggreg := true;
end;

function TZREvCountFunction.Value(FiFo : TZRFiFo) : TZREvResult;
begin
  GetArguments(FiFo);
  if (ArgList.Count = 0) then
  begin
    if FiFo.Aggreg then
      inc(FCount);
    Result.Kind := resInt;
    Result.intResult := FCount;
  end else
    Result.Kind := resError;
  FreeArguments;
end;

procedure TZREvCountFunction.Reset;
begin
  FCount := 0;
end;

{ TZREvElementDataField }

constructor TZREvElementDataField.CreateField(aField : TField);
begin
  inherited Create;
  FDataSet := aField.DataSet;
  FFieldNo := aField.Index;
  FField := aField;
end;

function TZREvElementDataField.Value(FiFo : TZRFiFo) : TZREvResult;
begin
  if FDataSet.DefaultFields then
    FField := FDataSet.Fields[FFieldNo];
  if FField is TStringField then
  begin
    result.Kind := resString;
    result.strResult := TStringField(FField).Value;
  end else
    if (FField is TIntegerField) or
       (FField is TSmallIntField) or
       (FField is TWordField) then
    begin
      result.Kind := resInt;
      result.intResult := FField.AsInteger;
    end else
      if (FField is TFloatField)  or
         (FField is TCurrencyField) or
         (FField is TBCDField) then
      begin
        result.Kind := resDouble;
        result.dblResult := TFloatField(FField).AsFloat;
      end else
        if FField is TBooleanField then
        begin
          result.Kind := resBool;
          result.BooResult := TBooleanField(FField).Value;
        end else
          if FField is TDateField then
          begin
            result.Kind := resString;
            result.strResult := TDateField(FField).AsString;
          end else
            if FField is TDateTimeField then
            begin
              result.Kind := resString;
              result.strResult := TDateField(FField).AsString;
            end else
              result.Kind := resError
end;

function TZREvElementError.Value(FiFo : TZRFiFo) : TZREvResult;
begin
  result.Kind := resError;
end;

{ TZREvaluator }

constructor TZREvaluator.Create;
begin
  Prepared := false;
end;

destructor TZREvaluator.Destroy;
begin
  if Prepared then Unprepare;
  inherited Destroy;
end;

procedure TZREvaluator.TrimString(var strString : string);
var
  intStart,
  intEnd : integer;
begin
  intStart := 1;
  intEnd := Length(strString);
  while (copy(strString, intStart,1) = ' ') and (intStart < intEnd) do
    inc(intStart);
  while (copy(strString, intEnd,1) = ' ') and (intEnd > intStart) do
    dec(intEnd);
  strString := Copy(strString, intStart, intEnd - intStart + 1);
end;

procedure TZREvaluator.FindDelimiter(strArg : string; var Pos : integer);
var
  n             : integer;
  FoundDelim    : boolean;
  booString     : boolean;
  intParenteses : integer;
begin
  if strArg='' then
    Pos := 0
  else
  begin
    FoundDelim := false;
    BooString := false;
    intParenteses := 0;
    N := 1;
    while (N<length(strArg)) and not FoundDelim do
    begin
      case StrArg[N] of
        '(' : if not booString then inc(intParenteses);
        ')' : if not booString then dec(intParenteses);
        '''' : booString := not booString;
      end;
      if (intParenteses=0) and not booString then
        if strArg[N]=ArgSeparator then
        begin
          FoundDelim := true;
          break;
        end;
      inc(N);
    end;
    if FoundDelim then
      Pos := N
    else
      Pos := 0;
  end;
end;

procedure TZREvaluator.FindOp1(const strExpr : string; var Op : TZREvOperator; var Pos, Len : integer);
var
  n : integer;
  booFound : boolean;
  intParenteses : integer;
  aString : string[255];
  booString : boolean;
begin
  n := 1;
  intParenteses := 0;
  booFound := false;
  Len := 1;
  aString := strExpr;
  booString := false;
  while (n < Length(strExpr)) and (not booFound) do
  begin
    booFound := true;
    case aString[N] of
      '(' : if not booString then inc(intParenteses);
      ')' : if not booString then dec(intParenteses);
      '''' : booString := not booString;
    end;
    if (intParenteses = 0) and (n > 1) and not booString then
      case aString[N] of
        '<' : begin
                if aString[N + 1] = '>' then
                begin
                  Op := opUnequal;
                  Len := 2;
                end else
                  if aString[N + 1] = '=' then
                  begin
                    Op := opLessOrEqual;
                    Len := 2;
                  end else
                    Op := opLess;
                end;
        '>' : if aString[N + 1] = '=' then
              begin
                Op := opGreaterOrEqual;
                Len := 2;
              end else
                Op := opGreater;
        '=' : Op := opEqual;
      else
        booFound := false;
      end
    else
      booFound := false;
    inc(N);
  end;
  if booFound then
    Pos := n - 1
  else
    Pos := -1;
end;

procedure TZREvaluator.FindOp2(const strExpr : string; var Op : TZREvOperator; var Pos, Len : integer);
var
  n : integer;
  booFound : boolean;
  intParenteses : integer;
  booString : boolean;
  aString : string[255];
begin
  n := 1;
  intParenteses := 0;
  booFound := false;
  booString := false;
  aString := strExpr;
  Len := 1;
  while (n < Length(strExpr)) and (not booFound) do
  begin
    booFound := true;
    case aString[N] of
      '(' : if not boostring then inc(intParenteses);
      ')' : if not boostring then dec(intParenteses);
      '''': booString := not booString;
    end;
    if (intParenteses = 0) and (not booString) and (N > 1) then
      case aString[N] of
        '+' : Op := opPlus;
        '-' : Op := opMinus;
        ' ' : if (AnsiLowercase(copy(strExpr, N + 1, 3)) = 'or ') then
              begin
                Op := opOr;
                Len := 2;
                inc(N);
              end else
                booFound := false;
      else
        booFound := false;
    end else
      booFound := false;
    inc(N);
  end;
  if booFound then
    Pos := N - 1
  else
    Pos := -1;
end;

procedure TZREvaluator.FindOp3(const strExpr : string; var Op : TZREvOperator; var Pos, Len : integer);
var
  n : integer;
  booFound : boolean;
  intParenteses : integer;
  booString : boolean;
  aString : string[255];
begin
  n := 1;
  intParenteses := 0;
  booFound := false;
  booString := false;
  Len := 1;
  aString := strExpr;
  while (N < Length(strExpr)) and (not booFound) do
  begin
    booFound := true;
    case aString[N] of
      '(' : if not booString then inc(intParenteses);
      ')' : if not booString then dec(intParenteses);
      '''': booString := not booString;
    end;
    if (intParenteses = 0) and (not booString) and (N > 1) then
    begin
      case aString[N] of
        '*' : Op := opMul;
        '/' : Op := opDiv;
        ' ' : if (AnsiLowercase(copy(strExpr, n + 1, 4)) = 'and ') then
              begin
                Op := opAnd;
                Len := 3;
                inc(N);
              end else
                booFound := false;
      else
        booFound := false;
    end;
  end else
    booFound := false;
    inc(N);
  end;
  if booFound then
    Pos := N - 1
  else
    Pos := -1;
end;

function TZREvaluator.NegateResult(const Res : TZREvResult) : TZREvResult;
begin
  result.Kind := Res.Kind;
  case Res.Kind of
    resInt: result.intResult := - Res.intResult;
    resDouble: result.dblResult := -Res.dblResult;
    resBool: result.booResult := not Res.booResult;
  else
    result.Kind := resError;
  end;
end;

function TZREvaluator.EvalVariable(strVariable : string) : TZREvResult;
var
  SeparatorPos : integer;
  DSName : string;
  FieldName : string;
  aDataSet : TDataSet;
  aField : TField;
  I : integer;
begin
  if assigned(FDataSets) then
  begin
    SeparatorPos := AnsiPos('.', strVariable);
    DSName := AnsiUpperCase(copy(StrVariable, 1, SeparatorPos - 1));
    FieldName := AnsiUpperCase(copy(strVariable, SeparatorPos + 1, length(StrVariable) - SeparatorPos));
    aField := nil;
    aDataSet := nil;
    if length(DSName)>0 then
    begin
      for I := 0 to FDataSets.Count - 1 do
        if AnsiUpperCase(TDataSet(FDataSets[I]).Name) = DSName then
        begin
          aDataSet := TDataSet(FDataSets[I]);
          break;
        end;
      if aDataSet<>nil then
        aField := aDataSet.FindField(FieldName);
    end else
    begin
      for I := 0 to FDataSets.Count - 1 do
        with TDataSet(FDataSets[I]) do
        begin
          aField := FindField(FieldName);
          if aField <> nil then break;
        end;
    end;
    if aField <> nil then
      FiFo.Put(TZREvElementDataField.CreateField(aField))
    else
      FiFo.Put(TZREvElementError.Create);
  end else
    FiFo.Put(TZREvElementError.Create);
end;

function TZREvaluator.EvalString(const strString : string) : TZREvResult;
begin
  result.Kind := resString;
  result.strResult := strString;
  FiFo.Put(TZREvElementString.CreateString(Result.StrResult));
end;

function TZREvaluator.EvalFunction(strFunc : string; const strArg : string) : TZREvResult;
var
  DelimPos : integer;
  aString : string;
  Res : TZREvResult;
  aFunc : TZREvElement;
begin
  StrFunc := AnsiUpperCase(StrFunc);
  aFunc := ZRFunctionLibrary.GetFunction(strFunc);
  if AFunc is TZREvElementError then
  begin
    if StrArg = '' then
    begin
      AFunc.Free;
      EvalVariable(StrFunc)
    end else
      FiFo.Put(AFunc);
  end else
  begin
    FiFo.Put(AFunc);
    if not (aFunc is TZREvElementError) then
    begin
      aString := strArg;
      repeat
        FindDelimiter(aString, DelimPos);
        if DelimPos > 0 then
          Res := Evaluate(copy(aString, 1, DelimPos - 1))
        else
          if length(aString) > 0 then Res := Evaluate(aString);
        Delete(aString, 1, DelimPos);
      until DelimPos = 0;
    end;
    FiFo.Put(TZREvElementArgumentEnd.Create);
  end;
end;

function TZREvaluator.EvalConstant(const strConstant : string) : TZREvResult;
var
  N : integer;
  aString : string[255];
begin
  N := 1;
  aString := strConstant;
  while (N <= Length(aString)) and  (aString[N] in ['0'..'9']) do
    inc(N);
  result.Kind := resInt;
  while ((N <= Length(aString)) and (aString[N] in ['0'..'9', '.', 'e', 'E', '+', '-'])) do
  begin
    inc(N);
    result.Kind := resDouble;
  end;
  if N - 1 <> Length(aString) then
    result.Kind := resError
  else
  begin
    if result.Kind = resInt then
      result.intResult := StrToInt(aString)
    else
      if result.Kind = resDouble then
      begin
        if DecimalSeparator <> '.' then
        begin
          while pos('.', aString) > 0 do
            aString[pos('.', aString)] := DecimalSeparator;
        end;
        result.dblResult := StrToFloat(aString);
      end;
  end;
  if result.Kind=resError then
    FiFo.Put(TZREvElementError.Create)
  else
    FiFo.Put(TZREvElementConstant.CreateConstant(Result));
end;

function TZREvaluator.EvalFunctionExpr(const strFunc : string) : TZREvResult;
var
  argRes : TZREvResult;
  po : integer;
begin
  po := AnsiPos('(', StrFunc);
  if po > 0 then
  begin
    if copy(StrFunc, length(StrFunc), 1) = ')' then
    begin
      result := EvalFunction(copy(StrFunc, 1, po - 1), copy(StrFunc, po + 1, length(strFunc) - po - 1));
    end else
    begin
      argRes.Kind := resError;
      result := EvalFunction('', '');
    end;
  end else
  begin
    argRes.Kind := resError;
    result := EvalFunction(StrFunc, '');
  end;
end;

function TZREvaluator.EvalFactor(strFactorExpr : string) : TZREvResult;
var
  aString : string[255];
begin
  TrimString(strFactorExpr);
  aString := strFactorExpr;
  if (AnsiLowerCase(Copy(strFactorExpr, 1, 3)) = 'not') then
    result := NegateResult(EvalFactor(Copy(strFactorExpr, 4, Length(strFactorExpr))))
  else
    case aString[1] of
      '(' : if strFactorExpr[Length(strFactorExpr)] = ')' then
              result := Evaluate(Copy(strFactorExpr, 2, Length(strFactorExpr) - 2))
            else
              result.Kind := resError;
      '-' : result := NegateResult(EvalFactor(Copy(strFactorExpr, 2, Length(strFactorExpr))));
      '+' : result := EvalFactor(Copy(strFactorExpr, 2, Length(strFactorExpr)));
      '0'..'9' : result := EvalConstant(strFactorExpr);
      '''' : if aString[Length(strFactorExpr)] = '''' then
               result := EvalString(Copy(strFactorExpr, 2, Length(strFactorExpr) - 2))
             else
             begin
               result.Kind := resError;
               FiFo.Put(TZREvElementError.Create);
             end;
      '[' : if aString[Length(strFactorExpr)] = ']' then
              result := EvalVariable(Copy(strFactorExpr, 2, Length(strFactorExpr) - 2))
            else
              result.Kind := resError;
      'A'..'Z', 'a'..'z' : result := EvalFunctionExpr(strFactorExpr);
    else
    begin
      FiFo.Put(TZREvElementError.Create);
      result.Kind := resError;
    end;
  end;
end;

function TZREvaluator.EvalSimpleExpr(const strSimplExpr : string) : TZREvResult;
var
  op : TZREvOperator;
  intStart,
  intLen : integer;
  Res1,
  Res2 : TZREvResult;
begin
  FindOp2(strSimplExpr, op, intStart, intLen);
  if intStart > 0 then
  begin
    FiFo.Put(TZREvElementOperator.CreateOperator(Op));
    Res1 := EvalTerm(Copy(strSimplExpr, 1, intStart - 1));
    Res2 := EvalSimpleExpr(Copy(strSimplExpr, intStart + intLen, Length(strSimplExpr)));
  end else
    result := EvalTerm(strSimplExpr);
end;

function TZREvaluator.EvalTerm(const strTermExpr : string) : TZREvResult;
var
  op : TZREvOperator;
  intStart,
  intLen : integer;
  Res1,
  Res2 : TZREvResult;
begin
  FindOp3(strTermExpr, op, intStart, intLen);
  if intStart > 0 then
  begin
    FiFo.Put(TZREvElementOperator.CreateOperator(Op));
    Res1 := EvalFactor(Copy(strTermExpr, 1, intStart - 1));
    Res2 := EvalTerm(Copy(strTermExpr, intStart + intLen, Length(strTermExpr)));
  end else
    result := EvalFactor(strTermExpr);
end;

function TZREvaluator.Evaluate(const strExpr : string) : TZREvResult;
var
  op : TZREvOperator;
  intStart,
  intLen : integer;
  Res1,
  Res2 : TZREvResult;
begin
  FindOp1(strExpr, op, intStart, intLen);
  if intStart > 0 then
  begin
    FiFo.Put(TZREvElementOperator.CreateOperator(Op));
    Res1 := EvalSimpleExpr(Copy(strExpr, 1, intStart - 1));
    Res2 := EvalSimpleExpr(Copy(strExpr, intStart + intLen, Length(strExpr)));
  end else
    result := EvalSimpleExpr(strExpr);
end;

procedure TZREvaluator.Prepare(const strExpr : string);
var
  Value : TZREvResult;
begin
  if Prepared then Unprepare;
  FiFo := TZRFiFo.Create;
  if strExpr = '' then
    Value := Evaluate(''' ''')
  else
    Value := Evaluate(strExpr);
  Prepared := true;
end;

procedure TZREvaluator.UnPrepare;
begin
  if Assigned(FiFo) then
    begin
      FiFo.Free;
      FiFo:= nil;
    end;
  Prepared := false;
end;

procedure TZREvaluator.Reset;
var
  I : integer;
begin
  for I := 0 to FiFo.FiFo.Count - 1 do
    TZREvElement(FiFo.FiFo[I]).Reset;
end;

function TZREvaluator.Value : TZREvResult;
begin
  FiFo.Start;
  Result := TZREvElement(FiFo.Get).Value(FiFo);
end;

function TZREvaluator.GetIsAggreg : boolean;
var
  I : integer;
begin
  Result := false;
  for I := 0 to FiFo.FiFo.Count - 1 do
    Result := Result or TZREvElement(FiFo.FiFo[I]).IsAggreg;
end;

function TZREvaluator.GetAggregate : boolean;
begin
  Result := FiFo.Aggreg;
end;

procedure TZREvaluator.SetAggregate(Value : boolean);
begin
  FiFo.Aggreg := Value;
end;

function TZREvaluator.Calculate(const strExpr : string) : TZREvResult;
begin
  Prepare(strExpr);
  result := Value;
  UnPrepare;
end;

initialization
  ZRFunctionLibrary := TZRFunctionLibrary.Create;

  { Register ZREvaluator functions }
  RegisterZRFunction(TZREvIfFunction,'IF', 'IF(<Exp>, <X>, <Y>)|' + LoadStr(SzrIfDesc), LoadStr(SzrQSD), '5BVV');
  RegisterZRFunction(TZREvStrFunction, 'STR', 'STR(<X>)|' + LoadStr(SzrStrDesc), LoadStr(SzrQSD), '7N');
  RegisterZRFunction(TZREvUpperFunction, 'UPPER', 'UPPER(<X>)|' + LoadStr(SzrUpperDesc), LoadStr(SzrQSD), '7S');
  RegisterZRFunction(TZREvLowerFunction, 'LOWER', 'LOWER(<X>)|' + LoadStr(SzrLowerDesc), LoadStr(SzrQSD), '7S');
  RegisterZRFunction(TZREvPrettyFunction, 'PRETTY', 'PRETTY(<X>)|' + LoadStr(SzrPrettyDesc), LoadStr(SzrQSD), '7S');
  RegisterZRFunction(TZREvTimeFunction, 'TIME', 'TIME|' + LoadStr(SzrTimeDesc), LoadStr(SzrQSD), '1');
  RegisterZRFunction(TZREvDateFunction, 'DATE', 'DATE|' + LoadStr(SzrDateDesc), LoadStr(SzrQSD), '1');
  RegisterZRFunction(TZREvCopyFunction, 'COPY', 'COPY(<X>, <St>,<Len>)|' + LoadStr(SzrCopyDesc), LoadStr(SzrQSD), '7SNN');
  RegisterZRFunction(TZREvSumFunction, 'SUM', 'SUM(<X>)|' + LoadStr(SzrSumDesc), LoadStr(SzrQSD), '3N');
  RegisterZRFunction(TZREvCountFunction, 'COUNT', 'COUNT|'+ LoadStr(SzrCountDesc), LoadStr(SzrQSD), '3');
  RegisterZRFunction(TZREvMaxFunction, 'MAX', 'MAX(<X>)|' + LoadStr(SzrMaxDesc), LoadStr(SzrQSD), '3V');
  RegisterZRFunction(TZREvMinFunction, 'MIN', 'MIN(<X>)|' + LoadStr(SzrMinDesc), LoadStr(SzrQSD), '3V');
  RegisterZRFunction(TZREvAverageFunction, 'AVERAGE', 'AVERAGE(<X>)|' + LoadStr(SzrAverageDesc), LoadStr(SzrQSD), '3N');
  RegisterZRFunction(TZREvTrue, 'TRUE', 'TRUE|' + LoadStr(SzrBoolDesc), LoadStr(SzrQSD), '5');
  RegisterZRFunction(TZREvFalse, 'FALSE', 'FALSE|' + LoadStr(SzrBoolDesc), LoadStr(SzrQSD), '5');
  RegisterZRFunction(TZREvIntFunction, 'INT', 'INT(<X>)|' + LoadStr(SzrIntDesc), LoadStr(SzrQSD), '2N');
  RegisterZRFunction(TZREvFracFunction, 'FRAC', 'FRAC(<X>)|' + LoadStr(SzrFracDesc), LoadStr(SzrQSD), '2N');
  RegisterZRFunction(TZREvSQRTFunction, 'SQRT', 'SQRT(<X>)|' + LoadStr(SzrSqrtDesc), LoadStr(SzrQSD), '2N');
  RegisterZRFunction(TZREvTypeOfFunction, 'TYPEOF', 'TYPEOF(<Exp>)|' + LoadStr(SzrTypeofDesc), LoadStr(SzrQSD), '6N');
  RegisterZRFunction(TZREvFormatNumericFunction,'FORMATNUMERIC', 'FORMATNUMERIC(<F>,<N>|' + LoadStr(SzrFormatNumericDesc), LoadStr(SzrQSD), '7SN');

finalization
  ZRFunctionLibrary.Free;

end.



type
  TZREvaluator = class
    private
      fExpression: String;
    protected
    public
      constructor Create; override;
      destructor  Destroy; override;
      procedure Prepare;
      property Expression: String read fExpression write fExpression default '';
  end;
  
constructor TZREvaluator.Create;
begin
  fExpression:= '';
end;

destructor TZREvaluator.Destroy;
begin
end;

procedure TZREvaluator.Prepare;
var
  Stream: TStringStream;
  Parser: TParser;
  
  procedure ProcessToken;
  begin
    case Parser.Token of
      toSymbol :
      toString :
      toInteger:
      toFloat  :
    end;
  end;
  
begin
  Stream:= TStringStream.Create(fExpression);
  Parser:= TParser.Create(Stream);
  try
    while Parser.Token <> toEOF do
      begin
        ProcessToken;
        Parser.NextToken;
      end;
  finally
    Parser.Free;
    Stream.Free;
  end;
end;


